/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/

 
package org.jabotics.robot.core;

import java.util.ArrayList;
import org.jabotics.robot.defs.*;
import org.jabotics.robot.util.*;
import org.jabotics.robot.compile.*;
import org.jabotics.robot.adapter.XAdapter;

/**
 * CEventDispatcher is responsible for sending sensor event messages to registered listeners.
 * This will be done in a dedicated dispatcher thread.
 * 
 * @author  Frank Gollnik
 * @version 1.1.2
 */
public class CEventDispatcher
{
    // class variables:
    static private final int MAX_QUEUE_SIZE = 10;
    static private final int UPDATE_DELAY = 4; // in ms
    static private final int DISPATCHER_THREAD_PRIORITY = Thread.MAX_PRIORITY - 4;
    static private final ITimerAdapter sTimer;
    static private final CEventDispatcher sDispatcher;

    // for session management (in conjunction with simulation module)
    static private volatile boolean sStopController = false;
    static private volatile boolean sControllerHalted = false;


    static
    {
        IAdapterFactory adapterFactory = XAdapter.getFactory();
        sTimer = adapterFactory.getTimerAdapter();
        sDispatcher = new CEventDispatcher();

        if (CS.SESSIONS_ENABLED)
        {
            ISessionAdapter session = adapterFactory.getSessionAdapter();
            session.add(new SessionDataManager());
        }
    }


    /**
     * <font color="Red"><b>&bull;</b></font> Diese Methode wird vom Ereignismelder aufgerufen, wenn sich dessen Zustand ändert.
     * Diese Methode muss so programmiert sein, dass sie in jedem Falle sehr schnell ausgeführt wird,
     * damit die Sensor-, Motor- und Bewegungssteuerung nicht beeinträchtigt wird.
     * 
     * @param   port       sensor port reporting event
     * @param   oldValue   old value or state
     * @param   newValue   new value or state
     */
    public interface IHandler
    {
        void manageEvent(int port_id, float oldValue, float newValue, float oldRate, float newRate);
    }


    static private class SessionDataManager implements ISessionData
    {
        public String getKey()
        {
            return "CEventDispatcher.SessionDataManager";
        }

        @Override
        public void stopSession()
        {
            try
            {
                sStopController = true;
                while (!sControllerHalted) Thread.sleep(1);
            }
            catch (Exception e) {}
        }

        @Override
        public ISessionData resetSession()
        {
            sDispatcher.initialize();
            return this;
        }

        @Override
        public void startSession()
        {
            sStopController = false;
            sControllerHalted = false; // do not wait for controller thread to update sControllerHalted
        }
    }


    static private class XEvent implements TICloneable<XEvent>
    {
        static final XEvent PROTOTYPE = new XEvent();

        volatile IHandler mHandler = null;
        volatile float mOldValue = 0f;
        volatile float mNewValue = 0f;
        volatile float mOldRate = 0f;
        volatile float mNewRate = 0f;

        @Override
        public void copyDataFrom(XEvent template)
        {
            mHandler  = template.mHandler;
            mOldValue = template.mOldValue;
            mNewValue = template.mNewValue;
            mOldRate = template.mOldRate;
            mNewRate = template.mNewRate;
        }

        @Override
        public XEvent getClone()
        {
            XEvent clone = new XEvent();
            clone.copyDataFrom(this);
            return clone;
        }
    }


    static CEventDispatcher getDispatcher()
    {
        return sDispatcher;
    }

    
    // Instanzvariablen
    private final XTimedThread mThread = new XTimedThread("Event Dispatcher Thread",new DispatcherRunnable());
    private final ArrayList<TPipe<XEvent>>    mEventPipe = new ArrayList<TPipe<XEvent>>(D.TOTAL_SENSOR_NUMBER);
    private final TVolatileRefArray<IHandler> mHandler = new TVolatileRefArray<IHandler>(D.TOTAL_SENSOR_NUMBER);
    private final TVolatileRefArray<IHandler> mNewHandler = new TVolatileRefArray<IHandler>(D.TOTAL_SENSOR_NUMBER);
    private final TVolatileRefArray<Boolean>  mOverflow = new TVolatileRefArray<Boolean>(D.TOTAL_SENSOR_NUMBER);
    private volatile boolean mSensorsUpdated;

    /**
     * Konstruktor für Objekte der Klasse CEventDispatcher
     */
    private CEventDispatcher()
    {
        initialize();

        mThread.setPriority(DISPATCHER_THREAD_PRIORITY);
        mThread.setDaemon(true);
        mThread.setTimeouts(UPDATE_DELAY,UPDATE_DELAY,UPDATE_DELAY);
        mThread.start();
    }

    private void initialize()
    {
        for (int port_id = 0;  port_id < D.TOTAL_SENSOR_NUMBER;  ++port_id)
        {
            mEventPipe.add(port_id, new TPipe<XEvent>(XEvent.PROTOTYPE));
            mHandler.set(port_id,null);
            mNewHandler.set(port_id,null);
            mOverflow.set(port_id,false);
            mSensorsUpdated = false;
        }
    }

    private class DispatcherRunnable implements XTimedThread.Runnable
    {

        @Override
        public void init()
        {
        }

        @Override
        public void runTimed()
        {
            if (CS.SESSIONS_ENABLED)
            {
                sControllerHalted = sStopController;
                if (sStopController) return;
            }

            if (mSensorsUpdated)
            {
                mSensorsUpdated = false;
                for (int port_id = 0;  port_id < D.TOTAL_SENSOR_NUMBER;  ++port_id)
                {
                    IHandler handler = mHandler.get(port_id);
                    IHandler newHandler = mNewHandler.get(port_id);
                    boolean hasNewHandler = (newHandler != handler);
 
                    TPipe<XEvent> pipe = mEventPipe.get(port_id);
                    while (pipe.hasNext() || hasNewHandler)
                    {
                        while (pipe.hasNext())
                        {
                            XEvent event = pipe.read();
                            event.mHandler.manageEvent(port_id, event.mOldValue, event.mNewValue, event.mOldRate, event.mNewRate);
                        }

                        if (hasNewHandler)
                        {
                            // new handler has been detected before event queue has been emptied
                            // => no events left for previous handler
                            mHandler.set(port_id,newHandler);
                            handler = newHandler;
                            //mOverflow.set(port_id,false);
                        }

                        // check if new handler has been set in the meantime
                        newHandler = mNewHandler.get(port_id);
                        hasNewHandler = (newHandler != handler);
                    }
                }
            }
        }

        @Override
        public void terminate()
        {
        }

        @Override
        public void shutdown()
        {
        }
    }

    // After call to this method, the previous event handler will for sure get no further event messages
    // and the new handler will have received an initial event notification
    void setEventHandler(ISensorAdapter adapter, IHandler handler)
    {
        // is synchronized on the calling sensor
        int port_id = adapter.getPortID();
        mNewHandler.set(port_id,handler);
        while (mHandler.get(port_id) != handler) sTimer.msDelay(1); // make sure, handler switch has been completed
    }

    // call to this method clears the overflow flag, if set
    boolean getAndResetOverflowFlag(ISensorAdapter adapter)
    {
        int port_id = adapter.getPortID();
        boolean overflow = mOverflow.get(port_id);
        if (overflow) mOverflow.set(port_id,false); // only clear the flag, if overflow for sure has been detected by this thread
        return overflow;
    }

    void addEvent(IHandler handler, ISensorAdapter adapter, float oldValue, float newValue, float oldRate, float newRate)
    {
        int port_id = adapter.getPortID();

        if (null == mNewHandler.get(port_id)) return;
        
        TPipe<XEvent> pipe = mEventPipe.get(port_id);
        if (pipe.getSize() >= MAX_QUEUE_SIZE)
        {
           mOverflow.set(port_id,true);
           return;
        }

        XEvent event = pipe.getWriteObject();
        event.mHandler = handler;
        event.mOldValue = oldValue;
        event.mNewValue = newValue;
        event.mOldRate = oldRate;
        event.mNewRate = newRate;
        pipe.commitWriteObject();
    }

    void notifyOnSensorUpdates()
    {
        mSensorsUpdated = true;
    }
}
