/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.en;


/**
 * <font color="Gold"><b>&bull;</b></font> The interface IObjectFactory allows to create
 * all object instances necessary to control the robot (like sensor and motor objects).
 * Users can access these objects via the provided interfaces.
 * 
 * @author  Frank Gollnik
 * @version 1.1.2
 */

public interface IObjectFactory
{
    /**
     * <font color="Gold"><b>&bull;</b></font> Creates an object instance representing a motor of the robot.
     * 
     * @param  port   Port, that the motor is connected to
     * @return        Interface of type IMotor to the new object instance
     */
    public IMotor createMotor(EMotorPort port);
    
    /**
     * <font color="Gold"><b>&bull;</b></font>  Creates an object instance representing an auto-pilot for controlling robot motions.
     * Default values for wheel diameter and wheelbase will be taken to initialize the pilot instance.
     * 
     * @param  portLeftMotor   Port, that the left motor is connected to
     * @param  portRightMotor  Port, that the right motor is connected to
     * @return                 Interface of type IPilot to the new object instance
     */
    public IPilot createPilot(EMotorPort portLeftMotor, EMotorPort portRightMotor);
    
    /**
     * <font color="Gold"><b>&bull;</b></font>  Creates an object instance representing an auto-pilot for controlling robot motions
     * with given values for wheel diameter and wheelbase of the robot.
     * 
     * @param  portLeftMotor   Port, that the left motor is connected to
     * @param  portRightMotor  Port, that the right motor is connected to
     * @param  wheelDiameter   wheel diameter of the robot (in cm)
     * @param  wheelbase       wheelbase of the robot (in cm)
     * @return                 Interface of type IPilot to the new object instance
     */
    public IPilot createPilot(EMotorPort portLeftMotor, EMotorPort portRightMotor, float wheelDiameter, float wheelbase);

    /**
     * <font color="Gold"><b>&bull;</b></font> Creates an object instance representing a button of the robot brick.
     * 
     * @param  button  Button type (ENTER, STOP, LEFT, RIGHT)
     * @return         Interface of type ITouchSensor to the new object instance
     */
    public ITouchSensor createTouchSensor(EButton button);

    /**
     * <font color="Gold"><b>&bull;</b></font> Creates an object instance representing a touch sensor connected to a sensor port.
     * 
     * @param  port   Port, that the sensor is connected to
     * @return        Interface of type ITouchSensor to the new object instance
     */
    public ITouchSensor createTouchSensor(ESensorPort port);
    
    /**
     * <font color="Gold"><b>&bull;</b></font> Creates an object instance representing a sound sensor connected to a sensor port.
     * 
     * @param  port   Port, that the sensor is connected to
     * @return        Interface of type ISensor to the new object instance
     */
    public ISensor createSoundSensor(ESensorPort port);
    
    /**
     * <font color="Gold"><b>&bull;</b></font> Creates an object instance representing a light sensor connected to a sensor port.
     * 
     * @param  port   Port, that the sensor is connected to
     * @return        Interface of type ILightSensor to the new object instance
     */
    public ILightSensor createLightSensor(ESensorPort port);
    
    /**
     * <font color="Gold"><b>&bull;</b></font> Creates an object instance representing an ultrasonic sensor connected to a sensor port.
     * 
     * @param  port   Port, that the sensor is connected to
     * @return        Interface of type ISensor to the new object instance
     */
    public ISensor createUltrasonicSensor(ESensorPort port);
    
    /**
     * <font color="Gold"><b>&bull;</b></font> Creates an object instance representing the LCD display of the robot brick.
     * 
     * @return        Interface of type IDisplay to the new object instance
     */
    public IDisplay createDisplay();

    /**
     * <font color="Gold"><b>&bull;</b></font> Creates an object instance representing the speaker of the robot brick.
     * 
     * @return        Interface of type ISpeaker to the new object instance
     */
    public ISpeaker createSpeaker();

    /**
     * <font color="Gold"><b>&bull;</b></font> Creates an object instance representing a timer.
     * 
     * @return        Interface of type ITimer to the new object instance
     */
    public ITimer createTimer();

    /**
     * <font color="Red"><b>&bull;</b></font> Creates an object instance representing a console window on a connected PC.
     * 
     * @return        Interface of type IPCConsole to the new object instance
     */
    public IPCConsole createPCConsole();
}
