/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.*;
import java.awt.geom.*;
import java.awt.Point;
import java.awt.Color;
import java.awt.GradientPaint;

/**
 * Beschreiben Sie hier die Klasse XLightSensor.
 * 
 * @author  Frank Gollnik
 * @version 1.1.2
 */
public class XLightSensor extends BSensor
{
    // Konstanten / Klassenvariablen:
    private static final Area DETECTION_AREA = new Area(new Rectangle2D.Float(0f,-0.1f,2000f,0.2f)); // in sensor's coordinate space
    private static XLightSensor sLightSensorDownward;
    private static XLightSensor sLightSensorForward;
    private static final Color PAINT_LIGHT_DOWNWARD_OFF = new Color(96,96,96);
    private static final GradientPaint PAINT_LIGHT_FORWARD_OFF;
    private static final GradientPaint PAINT_LIGHT_FORWARD_ON;
    static
    {
        Color c1 = new Color(0,0,0,32);
        Color c2 = new Color(0,0,0,96);
        PAINT_LIGHT_FORWARD_OFF = new GradientPaint(new Point.Float(-2f,0f), c1, new Point.Float(1f,0f), c2);
        Color c3 = new Color(255,0,0);
        Color c4 = new Color(255,0,0,64);
        PAINT_LIGHT_FORWARD_ON = new GradientPaint(new Point.Float(-2f,0f), c3, new Point.Float(1f,0f), c4);
    }



    static public synchronized XLightSensor getLightSensor(ESensorDirection direction)
    {
        if (null == sLightSensorDownward)
        {
            sLightSensorDownward = new XLightSensor(ESensorDirection.DOWNWARD);
            sLightSensorForward = new XLightSensor(ESensorDirection.FORWARD);
        }

        switch (direction)
        {
            case FORWARD: return sLightSensorForward;
            default:      return sLightSensorDownward;
        }
    }

    // Instanzvariablen:
    private double mNextUpdateTime = 0.0;
    private float mValueLightOff = 0;
    private float mValueLightOn = 0;
    private float mCenterX = 0f; // in robot's coordinate space
    private float mCenterY = 0f; // in robot's coordinate space
    private XGraphicsShape mGraphicsLED;
    private XFloor mFloor = null;
    private Area mDetectionArea = DETECTION_AREA; // in robot's coordinate space

    private XLightSensor(ESensorDirection direction)
    {
        super();
        mDirection = direction;
        buildGraphics();
    }

    private void buildGraphics()
    {
        if (mDirection == ESensorDirection.DOWNWARD)
        {
            mGraphicsLED = new XGraphicsShape(new Ellipse2D.Float(-0.7f,-0.7f,1.4f,1.4f));
        }
        else
        {
            Path2D.Float path = new Path2D.Float();
            path.moveTo(1f,-2.5f);
            path.quadTo(-2f,0f,1f,2.5f);
            path.closePath();
            mGraphicsLED = new XGraphicsShape(path);
        }
        mGraphicsLED.setDraw(false);
        this.updateGraphics();
        mGraphicsRep.add(mGraphicsLED);
    }

    @Override
    public void updateGraphics()
    {
        XSensorRef.RefData data = mLeadingRefData;
        mGraphicsLED.setFill(this.isReferenced());
        if (mDirection == ESensorDirection.DOWNWARD)
        {
            mGraphicsLED.setPaint(((null != data) && data.lightIsOn()) ? Color.red : PAINT_LIGHT_DOWNWARD_OFF);
        }
        else
        {
            mGraphicsLED.setPaint(((null != data) && data.lightIsOn()) ? PAINT_LIGHT_FORWARD_ON : PAINT_LIGHT_FORWARD_OFF);
        }
    }

    @Override
    public ESensorType getType()
    {
        return ESensorType.LIGHT_SENSOR;
    }

    @Override
    public synchronized float getValue(XSensorRef ref)
    {
        updateBrightness();
        return ref.lightIsOn() ? mValueLightOn : mValueLightOff;
    }

    @Override
    public boolean hasLight()
    {
        return true;
    }

    @Override
    public void manageVisitor(BModelVisitor visitor)
    {
        visitor.visitLightSensor(this);
    }

    @Override
    public synchronized void setTransform(float translationX, float translationY, float rotation)
    {
        mCenterX = translationX; // in robot's coordinate space
        mCenterY = translationY; // in robot's coordinate space
        AffineTransform atf = AffineTransform.getTranslateInstance(translationX,translationY);
        atf.rotate(rotation);
        mDetectionArea = DETECTION_AREA.createTransformedArea(atf); // in robot's coordinate space

        mGraphicsRep.setTransform(translationX, translationY, rotation);
    }

    void updateBrightness()
    {
        if (XSimulationModel.sTime >= mNextUpdateTime)
        {
            if (null == mFloor) mFloor = XSimulationModel.getScene().getFloor();

            if (mDirection == ESensorDirection.DOWNWARD)
            {
                XRobot.XDynamicData dd = XSimulationModel.getRobot().getDynamicData();
                float sin = (float)Math.sin(dd.getRobotPhi());
                float cos = (float)Math.cos(dd.getRobotPhi());
                float xSensor = (float)dd.getRobotX() + mCenterX * cos - mCenterY * sin;
                float ySensor = (float)dd.getRobotY() + mCenterX * sin + mCenterY * cos;
                mValueLightOn = mFloor.getMaxBrightness(xSensor,ySensor);
                mValueLightOff = mValueLightOn * mFloor.getLuminosity(xSensor,ySensor); // luminosity between 0f and 1f
            }
            else
            {
                float[] scene_coordinates = new float[2];
                float distance = getDistance(mDetectionArea, mCenterX, mCenterY, scene_coordinates);
                mValueLightOff = 100f * mFloor.getEdgeLuminosity(scene_coordinates[0],scene_coordinates[1]);
                float lum = 80f / ((distance < 1f) ? 1f : distance);
                //float lum = 120f / ((distance < 1f) ? 1f : distance * distance);
                //mValueLightOn = mValueLightOff + lum;
                mValueLightOn = (float) Math.sqrt(mValueLightOff*mValueLightOff + lum*lum);
                //float lum = 1f / ((distance < 1f) ? 1f : distance * distance);
                //mValueLightOn = (float)Math.log(Math.exp(mValueLightOff - 50) + lum) + 50;
            }

            if (mValueLightOn < 0)
            {
                mValueLightOn = 0f;
            }
            else if (mValueLightOn > 100f)
            {
                mValueLightOn = 100f;
            }

            if (mValueLightOff < 0)
            {
                mValueLightOff = 0f;
            }
            else if (mValueLightOff > 100f)
            {
                mValueLightOff = 100f;
            }

            mNextUpdateTime = XSimulationModel.sTime + 0.006;
        }
    }

    @Override
    public ISessionData resetSession()
    {
        mNextUpdateTime = 0.0;
        mValueLightOff = 0;
        mValueLightOn = 0;
        return this;
    }
}
