/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import javax.swing.*;
//import javax.swing.border.LineBorder;
//import java.util.concurrent.atomic.AtomicBoolean;


/**
 * Beschreiben Sie hier die Klasse XSpriteView.
 * 
 * @author  Frank Gollnik
 * @version 1.1.2
 */
public class XSpriteView extends JComponent
{
    /* Definitions
     * 
     * Coordinate systems (all "left handed", i.e. x to the right, y downwards):
     * - "physical coordinates": real world coordinates of the background area (e.g. in cm);
     *                           origin not necessarily in upper left corner:
     *                           position of origin according to definition of background rectangle
     * - "pixel coordinates":    real world coordinates scaled to pixel size
     *                           => (0|0) still does not correspond to upper left corner of the
     *                           background area
     * - "image coordinates":    shifted pixel coordinates so that origin (0|0) corresponds to
     *                           upper left corner of the background area
     * - "screen coordinates":   in pixels relative to upper left corner of current screen
     */

    static final AffineTransform AFT_IDENTITY = new AffineTransform();

    private class ImageData
    {
        private float mScale;         // in pixel/cm
        private BufferedImage mImage;  // (image coordinates)
        private BufferedImage mSprite; // (image coordinates)
        private Dimension mImageDim;   // in pixel
        private Dimension mSpriteDim;  // in pixel
    }

    private class BackgroundImageUpdater extends SwingWorker<ImageData, Object>
    {
        @Override
        public ImageData doInBackground()
        {
            return getImageData(mNewScale);
        }
 
        @Override
        protected void done()
        {
            try
            {
                mNewImageData = get();
                repaint();
                if (mNewScale == mNewImageData.mScale) mUpdater = null;
                if (this == mUpdater)
                {
                    // mNewScale has been changed in the meantime without executing an updater
                    mUpdater = new BackgroundImageUpdater(); // SwingWorker designed to be executed only once => create new one
                    mUpdater.execute();
                }
            }
            catch (Exception ignore)
            {
            }
        }
    }


    // instance variables:
    private Rectangle.Float mBackgroundRect;         // physical coordinates
    private Rectangle.Float mSpriteRect;             // physical coordinates relative to sprite position
    private volatile Point.Float mSpritePos;         // physical coordinates
    private volatile Point mSpriteUpperLeft;         // image coordinates
    private volatile Point mSpriteUpperLeftPainted;  // image coordinates for last painted sprite
    private volatile Point.Float mViewCenter;        // physical coordinates
    private volatile Point.Float mNewViewCenter;     // physical coordinates
    private volatile Rectangle.Float mPhysViewRect;  // current view rectangle in physical coordinates (e.g. in cm)
    private volatile ImageData mImageData = null;
    private volatile ImageData mNewImageData = null;
    private volatile float mNewScale;
    private volatile BackgroundImageUpdater mUpdater;
    private volatile boolean mRepaintSprite = true;
    private volatile Dimension mSize;
    private volatile int mHalfWidth;
    private volatile int mHalfHeight;
    private volatile int mImageToSreenX;
    private volatile int mImageToSreenY;

    // Constructors:
    protected XSpriteView()
    {
        super();
        super.setDoubleBuffered(false); // This component has its own buffering mechanism that makes double buffering obsolete.
    }

    protected void init(Rectangle.Float backgroundRect, Rectangle.Float spriteRect, float scale, Point.Float viewCenter, Point.Float spritePosition)
    {
        mBackgroundRect = backgroundRect;
        if (spriteRect.width > backgroundRect.width) spriteRect.width = backgroundRect.width;
        if (spriteRect.height > backgroundRect.height) spriteRect.height = backgroundRect.height;
        mSpriteRect = spriteRect;
        mViewCenter = viewCenter;
        mNewViewCenter = viewCenter;
        if (scale <= 0) scale = 1f;
        mImageData = getImageData(scale);
        mNewImageData = mImageData;
        mNewScale = scale;
        updateTransformationData();
        setSpritePosition(spritePosition);
        mSpriteUpperLeftPainted = mSpriteUpperLeft;
    }

    @Override
    final public void setDoubleBuffered(boolean aFlag)
    {
        // This component has its own buffering mechanism that makes double buffering obsolete.
    }

    @Override
    final public boolean isOpaque()
    {
        // This component is completely opaque by design.
        return true;
    }

    public void setViewCenter(Point.Float viewCenter)
    {
        mNewViewCenter = viewCenter;
        if (!mViewCenter.equals(mNewViewCenter)) repaint();
    }

    public Point.Float getViewCenter()
    {
        return mViewCenter;
    }

    public void setScale(float scale)
    {
        if ((scale > 0.) && (scale != mNewScale))
        {
            mNewScale = scale;
            BackgroundImageUpdater updater = new BackgroundImageUpdater();
            if (null == mUpdater)  mUpdater = updater;
            if (updater == mUpdater) updater.execute();
        }
    }

    public float getScale()
    {
        return mImageData.mScale;
    }

    private void setSpritePosition(Point.Float position)
    {
        // internal version does not trigger repaintSprite()
        float scale = mImageData.mScale;
        int upperLeftX = (int)((mSpriteRect.x + position.x - mBackgroundRect.x) * scale); // image coordinates
        int upperLeftY = (int)((mSpriteRect.y + position.y - mBackgroundRect.y) * scale); // image coordinates
        mSpritePos = position;
        mSpriteUpperLeft = new Point(upperLeftX - 2, upperLeftY - 2);
    }

    public void setSpritePosition(Point.Float position, boolean forceRepaintSprite)
    {
        setSpritePosition(position);
        if (forceRepaintSprite || !mSpriteUpperLeft.equals(mSpriteUpperLeftPainted)) repaintSprite();
    }

    protected void repaintSprite()
    {
        mRepaintSprite = true;
        Rectangle rect1 = new Rectangle(mSpriteUpperLeftPainted.x + mImageToSreenX, mSpriteUpperLeftPainted.y + mImageToSreenY, mImageData.mSpriteDim.width, mImageData.mSpriteDim.height);
        Rectangle rect2 = new Rectangle(mSpriteUpperLeft.x + mImageToSreenX, mSpriteUpperLeft.y + mImageToSreenY, mImageData.mSpriteDim.width, mImageData.mSpriteDim.height);
        repaint(rect1.union(rect2));
    }

    private void updateTransformationData()
    {
        mSize = getSize();
        mHalfWidth = mSize.width / 2;
        mHalfHeight = mSize.height /2;
        mImageToSreenX = mHalfWidth + (int)((mBackgroundRect.x - mViewCenter.x) * mImageData.mScale);
        mImageToSreenY = mHalfHeight + (int)((mBackgroundRect.y - mViewCenter.y) * mImageData.mScale);
    }

    private void calcPhysicalViewRect()
    {
        float scale = mImageData.mScale;
        float x = mViewCenter.x - (float)mHalfWidth / scale;
        float y = mViewCenter.y - (float)mHalfHeight / scale;
        mPhysViewRect = new Rectangle.Float(x, y, (float)mSize.width / scale,  (float)mSize.height / scale);
    }

    public Rectangle2D.Float getPhysicalViewRect()
    {
        if (null == mPhysViewRect) calcPhysicalViewRect();
        return mPhysViewRect;
    }

    protected Point.Float transformScreenPosToPhysical(int x, int y)
    {
        float scale = mImageData.mScale;
        float dx = (float) (x - mHalfWidth);
        float dy = (float) (y - mHalfHeight);
        return new Point.Float(mViewCenter.x + dx / scale, mViewCenter.y + dy / scale);
    }

    protected Point transformPhysicalPosToScreen(float x, float y)
    {
        float scale = mImageData.mScale;
        int dx = (int) ((x - mViewCenter.x) * scale);
        int dy = (int) ((y - mViewCenter.y) * scale);
        return new Point(mHalfWidth + dx, mHalfHeight + dy);
    }

    /*private Rectangle getEnclosingSpriteRectScreen()
    {
        Dimension dim = mImageData.mSpriteDim;
        Point pos = mSpriteUpperLeft;
        return new Rectangle(pos.x + mImageToSreenX, pos.y + mImageToSreenY, dim.width, dim.height);
    }*/
    
    @Override
    final public void paintComponent(Graphics g)
    {
        super.paintComponent(g);

        Graphics2D g2D = (Graphics2D)g;
        AffineTransform aftInitial = g2D.getTransform();
        Paint paint = g2D.getPaint();
        g2D.setPaint(this.getBackground());

        if (mImageData != mNewImageData)
        {
            mImageData = mNewImageData;
            setSpritePosition(mSpritePos); // recalculates mSpriteUpperLeft
            mRepaintSprite = false;
            mViewCenter = mNewViewCenter;
            updateTransformationData();
        }
        else if ((mViewCenter != mNewViewCenter) || !mSize.equals(getSize()))
        {
            mViewCenter = mNewViewCenter;
            updateTransformationData();
        }

        float scale = mImageData.mScale;

        if (mRepaintSprite)
        {
            mRepaintSprite = false;
            restoreBackground();
            mSpriteUpperLeftPainted = mSpriteUpperLeft;
            saveBackground();
            Graphics2D g2DImage = mImageData.mImage.createGraphics();
            AffineTransform aft = g2DImage.getTransform();
            float dx = -scale * mBackgroundRect.x;
            float dy = -scale * mBackgroundRect.y;
            AffineTransform aftPhysical = new AffineTransform(scale, 0f, 0f, scale, dx, dy);
            g2DImage.setTransform(aftPhysical);
            // set clip to make sure that image will not be corrupted
            Rectangle.Float clip = new Rectangle.Float(mSpriteRect.x + mSpritePos.x, mSpriteRect.y + mSpritePos.y, mSpriteRect.width, mSpriteRect.height);
            g2DImage.setClip(clip);
            paintSprite(g2DImage);
            g2DImage.setTransform(aft);
        }

        Rectangle backgroundRect = new Rectangle(mImageData.mImageDim);   // transform image coordinates ...
        backgroundRect.translate(mImageToSreenX, mImageToSreenY);         // ... to screen coordinates

        Rectangle clip = g2D.getClipBounds();
        Rectangle clipOverlap = backgroundRect.intersection(clip);
        if (clipOverlap.isEmpty())
        {
            g2D.fill(clip);
        }
        else
        {
            Point ptClip = new Point(clip.x + clip.width, clip.y + clip.height); // lower right corner of clip
            Point ptClipOverlap = new Point(clipOverlap.x + clipOverlap.width, clipOverlap.y + clipOverlap.height); // lower right corner of clip overlap
            if (clipOverlap.x > clip.x) g2D.fill(new Rectangle(clip.x, clipOverlap.y, clipOverlap.x - clip.x, clipOverlap.height));
            if (ptClipOverlap.x < ptClip.x) g2D.fill(new Rectangle(ptClipOverlap.x, clipOverlap.y, ptClip.x - ptClipOverlap.x, clipOverlap.height));
            if (clipOverlap.y > clip.y) g2D.fill(new Rectangle(clip.x, clip.y, clip.width, clipOverlap.y - clip.y));
            if (ptClipOverlap.y < ptClip.y) g2D.fill(new Rectangle(clip.x, ptClipOverlap.y, clip.width, ptClip.y - ptClipOverlap.y));

            Rectangle rectSubimage = new Rectangle(clipOverlap);       // transform screen coordinates ...
            rectSubimage.translate(-mImageToSreenX, -mImageToSreenY);  // ... to image coordinates
            g2D.drawImage(mImageData.mImage.getSubimage(rectSubimage.x, rectSubimage.y, rectSubimage.width, rectSubimage.height), null, clipOverlap.x, clipOverlap.y);
        }

        g2D.setPaint(paint);
        g2D.setTransform(aftInitial);

        calcPhysicalViewRect();
    }

    private void restoreBackground()
    {
        if (null != mImageData.mSprite)
        {
            Graphics2D g2D = mImageData.mImage.createGraphics();
            AffineTransform aft = g2D.getTransform();
            g2D.setTransform(AFT_IDENTITY);
            g2D.drawImage(mImageData.mSprite, null, mSpriteUpperLeftPainted.x, mSpriteUpperLeftPainted.y);
            g2D.setTransform(aft);
        }
    }

    private void saveBackground()
    {
        if (null == mImageData.mSprite)
        {
            mImageData.mSprite = new BufferedImage(mImageData.mSpriteDim.width, mImageData.mSpriteDim.height, BufferedImage.TYPE_INT_RGB);
            mImageData.mSprite.setAccelerationPriority(1); // highest priority for acceleration (e.g keep in video memory if possible)
        }
        Graphics2D g2D = mImageData.mSprite.createGraphics();
        AffineTransform aft = g2D.getTransform();
        g2D.setTransform(AFT_IDENTITY);
        g2D.drawImage(mImageData.mImage.getSubimage(mSpriteUpperLeftPainted.x, mSpriteUpperLeftPainted.y, mImageData.mSpriteDim.width, mImageData.mSpriteDim.height), null, 0, 0);
        g2D.setTransform(aft);
    }

    private ImageData getImageData(float scale)
    {
        ImageData imgData = new ImageData();

        imgData.mScale = scale;

        int imageWidth = (int)(scale * mBackgroundRect.width);
        int imageHeight = (int)(scale * mBackgroundRect.height);
        imgData.mImageDim = new Dimension(imageWidth, imageHeight);
        imgData.mImage = new BufferedImage(imageWidth, imageHeight, BufferedImage.TYPE_INT_RGB);
        imgData.mImage.setAccelerationPriority(1); // highest priority for acceleration (e.g keep in video memory if possible)

        int spriteWidth = (int)(scale * mSpriteRect.width);
        int spriteHeight = (int)(scale * mSpriteRect.height);
        imgData.mSpriteDim = new Dimension(spriteWidth + 4, spriteHeight + 4);

        float dx = -scale * mBackgroundRect.x;
        float dy = -scale * mBackgroundRect.y;
        AffineTransform aftPhysical = new AffineTransform(scale, 0., 0., scale, dx, dy);

        Graphics2D g2D = imgData.mImage.createGraphics();
        g2D.setTransform(aftPhysical);
        paintBackgroundImage(g2D);
        
        return imgData;
    }

    protected void paintBackgroundImage(Graphics2D g2D)
    {
    }

    protected void paintSprite(Graphics2D g2D)
    {
    }
}
